@@ -19,6 +19,7 @@ import socket
 import logging
 import netaddr
 
+from ryu.ofproto import ether
 from ryu.ofproto import inet
 from ryu.ofproto import ofproto_v1_3
 from ryu.ofproto import ofproto_v1_3_parser
@@ -42,6 +43,78 @@ def str_to_int(src):
     return dst
 
 
+def to_action(dp, dic):
+    ofp = dp.ofproto
+    parser = dp.ofproto_parser
+
+    result = None
+    action_type = dic.get('type')
+    if action_type == 'OUTPUT':
+        out_port = int(dic.get('port', ofp.OFPP_ANY))
+        max_len = int(dic.get('max_len', ofp.OFPCML_MAX))
+        result = parser.OFPActionOutput(out_port, max_len)
+    elif action_type == 'COPY_TTL_OUT':
+        result = parser.OFPActionCopyTtlOut()
+    elif action_type == 'COPY_TTL_IN':
+        result = parser.OFPActionCopyTtlIn()
+    elif action_type == 'SET_MPLS_TTL':
+        mpls_ttl = int(dic.get('mpls_ttl'))
+        result = parser.OFPActionSetMplsTtl(mpls_ttl)
+    elif action_type == 'DEC_MPLS_TTL':
+        result = parser.OFPActionDecMplsTtl()
+    elif action_type == 'PUSH_VLAN':
+        ethertype = int(dic.get('ethertype'))
+        result = parser.OFPActionPushVlan(ethertype)
+    elif action_type == 'POP_VLAN':
+        result = parser.OFPActionPopVlan()
+    elif action_type == 'PUSH_MPLS':
+        ethertype = int(dic.get('ethertype'))
+        result = parser.OFPActionPushMpls(ethertype)
+    elif action_type == 'POP_MPLS':
+        ethertype = int(dic.get('ethertype'))
+        result = parser.OFPActionPopMpls(ethertype)
+    elif action_type == 'SET_QUEUE':
+        queue_id = int(dic.get('queue_id'))
+        result = parser.OFPActionSetQueue(queue_id)
+    elif action_type == 'GROUP':
+        group_id = int(dic.get('group_id'))
+        result = parser.OFPActionGroup(group_id)
+    elif action_type == 'SET_NW_TTL':
+        nw_ttl = int(dic.get('nw_ttl'))
+        result = parser.OFPActionSetNwTtl(nw_ttl)
+    elif action_type == 'DEC_NW_TTL':
+        result = parser.OFPActionDecNwTtl()
+    elif action_type == 'SET_FIELD':
+        field = dic.get('field')
+        value = dic.get('value')
+        if field == 'eth_dst':
+            field = ofp.OXM_OF_ETH_DST
+            value = mac.haddr_to_bin(str(value))
+        elif field == 'eth_src':
+            field = ofp.OXM_OF_ETH_SRC
+            value = mac.haddr_to_bin(str(value))
+        elif field == 'vlan_vid':
+            field = ofp.OXM_OF_VLAN_VID
+            value = int(value)
+        elif field == 'mpls_label':
+            field = ofp.OXM_OF_MPLS_LABEL
+            value = int(value)
+        else:
+            LOG.debug('Unknown field: %s' % field)
+            return None
+        f = parser.OFPMatchField.make(field, value)
+        result = parser.OFPActionSetField(f)
+    elif action_type == 'PUSH_PBB':
+        ethertype = int(dic.get('ethertype'))
+        result = parser.OFPActionPushPbb(ethertype)
+    elif action_type == 'POP_PBB':
+        result = parser.OFPActionPopPbb()
+    else:
+        LOG.debug('Unknown action type: %s' % action_type)
+
+    return result
+
+
 def to_actions(dp, acts):
     inst = []
     actions = []
@@ -49,88 +122,83 @@ def to_actions(dp, acts):
     parser = dp.ofproto_parser
 
     for a in acts:
-        action_type = a.get('type')
-        if action_type == 'OUTPUT':
-            out_port = int(a.get('port', ofproto_v1_3.OFPP_ANY))
-            max_len = int(a.get('max_len', ofproto_v1_3.OFPCML_MAX))
-            actions.append((parser.OFPActionOutput(out_port,
-                                                   max_len)))
-        elif action_type == 'COPY_TTL_OUT':
-            actions.append(parser.OFPActionCopyTtlOut())
-        elif action_type == 'COPY_TTL_IN':
-            actions.append(parser.OFPActionCopyTtlIn())
-        elif action_type == 'SET_MPLS_TTL':
-            mpls_ttl = int(a.get('mpls_ttl'))
-            actions.append((parser.OFPActionSetMplsTtl(mpls_ttl)))
-        elif action_type == 'DEC_MPLS_TTL':
-            actions.append((parser.OFPActionDecMplsTtl()))
-        elif action_type == 'PUSH_VLAN':
-            ethertype = int(a.get('ethertype'))
-            actions.append((parser.OFPActionPushVlan(ethertype)))
-        elif action_type == 'POP_VLAN':
-            actions.append(parser.OFPActionPopVlan())
-        elif action_type == 'PUSH_MPLS':
-            ethertype = int(a.get('ethertype'))
-            actions.append(parser.OFPActionPushMpls(ethertype))
-        elif action_type == 'POP_MPLS':
-            ethertype = int(a.get('ethertype'))
-            actions.append(parser.OFPActionPopMpls(ethertype))
-        elif action_type == 'SET_QUEUE':
-            queue_id = int(a.get('queue_id'))
-            actions.append(parser.OFPActionSetQueue(queue_id))
-        elif action_type == 'GROUP':
-            pass
-        elif action_type == 'SET_NW_TTL':
-            nw_ttl = int(a.get('nw_ttl'))
-            actions.append(parser.OFPActionSetNwTtl(nw_ttl))
-        elif action_type == 'DEC_NW_TTL':
-            actions.append(parser.OFPActionDecNwTtl())
-        elif action_type == 'SET_FIELD':
-            field = a.get('field')
-            value = a.get('value')
-            if field == 'eth_dst':
-                field = ofp.OXM_OF_ETH_DST
-                value = mac.haddr_to_bin(str(value))
-            elif field == 'eth_src':
-                field = ofp.OXM_OF_ETH_SRC
-                value = mac.haddr_to_bin(str(value))
-            elif field == 'vlan_vid':
-                field = ofp.OXM_OF_VLAN_VID
-                value = int(value)
-            elif field == 'mpls_label':
-                field = ofp.OXM_OF_MPLS_LABEL
-                value = int(value)
-            else:
-                LOG.debug('Unknown field: %s' % field)
-                continue
-            f = parser.OFPMatchField.make(field, value)
-            actions.append(parser.OFPActionSetField(f))
-        elif action_type == 'PUSH_PBB':
-            ethertype = int(a.get('ethertype'))
-            actions.append(parser.OFPActionPushPbb(ethertype))
-        elif action_type == 'POP_PBB':
-            actions.append(parser.OFPActionPopPbb())
-        elif action_type == 'GOTO_TABLE':
-            table_id = int(a.get('table_id'))
-            inst.append(parser.OFPInstructionGotoTable(table_id))
-        elif action_type == 'WRITE_METADATA':
-            metadata = str_to_int(a.get('metadata'))
-            metadata_mask = (str_to_int(a['metadata_mask'])
-                             if 'metadata_mask' in a
-                             else ofproto_v1_3_parser.UINT64_MAX)
-            inst.append(
-                parser.OFPInstructionWriteMetadata(metadata, metadata_mask))
-        elif action_type == 'METER':
-            meter_id = int(a.get('meter_id'))
-            inst.append(parser.OFPInstructionMeter(meter_id))
+        action = to_action(dp, a)
+        if action is not None:
+            actions.append(action)
         else:
-            LOG.debug('Unknown action type: %s' % action_type)
+            action_type = a.get('type')
+            if action_type == 'GOTO_TABLE':
+                table_id = int(a.get('table_id'))
+                inst.append(parser.OFPInstructionGotoTable(table_id))
+            elif action_type == 'WRITE_METADATA':
+                metadata = str_to_int(a.get('metadata'))
+                metadata_mask = (str_to_int(a['metadata_mask'])
+                                 if 'metadata_mask' in a
+                                 else parser.UINT64_MAX)
+                inst.append(
+                    parser.OFPInstructionWriteMetadata(
+                        metadata, metadata_mask))
+            elif action_type == 'METER':
+                meter_id = int(a.get('meter_id'))
+                inst.append(parser.OFPInstructionMeter(meter_id))
+            else:
+                LOG.debug('Unknown action type: %s' % action_type)
 
     inst.append(parser.OFPInstructionActions(ofp.OFPIT_APPLY_ACTIONS,
                                              actions))
     return inst
 
 
+def action_to_str(act):
+    action_type = act.cls_action_type
+
+    if action_type == ofproto_v1_3.OFPAT_OUTPUT:
+        buf = 'OUTPUT:' + str(act.port)
+    elif action_type == ofproto_v1_3.OFPAT_COPY_TTL_OUT:
+        buf = 'COPY_TTL_OUT'
+    elif action_type == ofproto_v1_3.OFPAT_COPY_TTL_IN:
+        buf = 'COPY_TTL_IN'
+    elif action_type == ofproto_v1_3.OFPAT_SET_MPLS_TTL:
+        buf = 'SET_MPLS_TTL:' + str(act.mpls_ttl)
+    elif action_type == ofproto_v1_3.OFPAT_DEC_MPLS_TTL:
+        buf = 'DEC_MPLS_TTL'
+    elif action_type == ofproto_v1_3.OFPAT_PUSH_VLAN:
+        buf = 'PUSH_VLAN:' + str(act.ethertype)
+    elif action_type == ofproto_v1_3.OFPAT_POP_VLAN:
+        buf = 'POP_VLAN'
+    elif action_type == ofproto_v1_3.OFPAT_PUSH_MPLS:
+        buf = 'PUSH_MPLS:' + str(act.ethertype)
+    elif action_type == ofproto_v1_3.OFPAT_POP_MPLS:
+        buf = 'POP_MPLS'
+    elif action_type == ofproto_v1_3.OFPAT_SET_QUEUE:
+        buf = 'SET_QUEUE:' + str(act.queue_id)
+    elif action_type == ofproto_v1_3.OFPAT_GROUP:
+        pass
+    elif action_type == ofproto_v1_3.OFPAT_SET_NW_TTL:
+        buf = 'SET_NW_TTL:' + str(act.nw_ttl)
+    elif action_type == ofproto_v1_3.OFPAT_DEC_NW_TTL:
+        buf = 'DEC_NW_TTL'
+    elif action_type == ofproto_v1_3.OFPAT_SET_FIELD:
+        if isinstance(act.field, ofproto_v1_3_parser.MTEthDst):
+            field = 'eth_dst'
+        elif isinstance(act.field, ofproto_v1_3_parser.MTEthSrc):
+            field = 'eth_src'
+        elif isinstance(act.field, ofproto_v1_3_parser.MTVlanVid):
+            field = 'vlan_vid'
+        elif isinstance(act.field, ofproto_v1_3_parser.MTMplsLabel):
+            field = 'mpls_label'
+        else:
+            field = '(Unknown field)'
+        buf = 'SET_FIELD: {%s:%s}' % (field, act.value)
+    elif action_type == ofproto_v1_3.OFPAT_PUSH_PBB:
+        buf = 'PUSH_PBB:' + str(act.ethertype)
+    elif action_type == ofproto_v1_3.OFPAT_POP_PBB:
+        buf = 'POP_PBB'
+    else:
+        buf = 'UNKNOWN'
+    return buf
+
+
 def actions_to_str(instructions):
     actions = []
 
@@ -138,43 +206,7 @@ def actions_to_str(instructions):
         if isinstance(instruction,
                       ofproto_v1_3_parser.OFPInstructionActions):
             for a in instruction.actions:
-                action_type = a.cls_action_type
-
-                if action_type == ofproto_v1_3.OFPAT_OUTPUT:
-                    buf = 'OUTPUT:' + str(a.port)
-                elif action_type == ofproto_v1_3.OFPAT_COPY_TTL_OUT:
-                    buf = 'COPY_TTL_OUT'
-                elif action_type == ofproto_v1_3.OFPAT_COPY_TTL_IN:
-                    buf = 'COPY_TTL_IN'
-                elif action_type == ofproto_v1_3.OFPAT_SET_MPLS_TTL:
-                    buf = 'SET_MPLS_TTL:' + str(a.mpls_ttl)
-                elif action_type == ofproto_v1_3.OFPAT_DEC_MPLS_TTL:
-                    buf = 'DEC_MPLS_TTL'
-                elif action_type == ofproto_v1_3.OFPAT_PUSH_VLAN:
-                    buf = 'PUSH_VLAN:' + str(a.ethertype)
-                elif action_type == ofproto_v1_3.OFPAT_POP_VLAN:
-                    buf = 'POP_VLAN'
-                elif action_type == ofproto_v1_3.OFPAT_PUSH_MPLS:
-                    buf = 'PUSH_MPLS:' + str(a.ethertype)
-                elif action_type == ofproto_v1_3.OFPAT_POP_MPLS:
-                    buf = 'POP_MPLS'
-                elif action_type == ofproto_v1_3.OFPAT_OFPAT_SET_QUEUE:
-                    buf = 'SET_QUEUE:' + str(a.queue_id)
-                elif action_type == ofproto_v1_3.OFPAT_GROUP:
-                    pass
-                elif action_type == ofproto_v1_3.OFPAT_SET_NW_TTL:
-                    buf = 'SET_NW_TTL:' + str(a.nw_ttl)
-                elif action_type == ofproto_v1_3.OFPAT_DEC_NW_TTL:
-                    buf = 'DEC_NW_TTL'
-                elif action_type == ofproto_v1_3.OFPAT_SET_FIELD:
-                    buf = 'SET_FIELD: {%s:%s}' % (a.field, a.value)
-                elif action_type == ofproto_v1_3.OFPAT_PUSH_PBB:
-                    buf = 'PUSH_PBB:' + str(a.ethertype)
-                elif action_type == ofproto_v1_3.OFPAT_POP_PBB:
-                    buf = 'POP_PBB'
-                else:
-                    buf = 'UNKNOWN'
-                actions.append(buf)
+                actions.append(action_to_str(a))
 
         elif isinstance(instruction,
                         ofproto_v1_3_parser.OFPInstructionGotoTable):
@@ -226,6 +258,8 @@ def to_match(dp, attrs):
                'tcp_dst': int,
                'udp_src': int,
                'udp_dst': int,
+               'arp_spa': to_match_ip,
+               'arp_tpa': to_match_ip,
                'ipv6_src': to_match_ipv6,
                'ipv6_dst': to_match_ipv6}
 
@@ -252,15 +286,27 @@ def to_match(dp, attrs):
                     'tcp_dst': to_match_tpdst,
                     'udp_src': to_match_tpsrc,
                     'udp_dst': to_match_tpdst,
+                    'arp_spa': match.set_arp_spa_masked,
+                    'arp_tpa': match.set_arp_tpa_masked,
                     'ipv6_src': match.set_ipv6_src_masked,
                     'ipv6_dst': match.set_ipv6_dst_masked}
 
+    if attrs.get('dl_type') == ether.ETH_TYPE_ARP or \
+            attrs.get('eth_type') == ether.ETH_TYPE_ARP:
+        if 'nw_src' in attrs and not 'arp_spa' in attrs:
+            attrs['arp_spa'] = attrs['nw_src']
+            del attrs['nw_src']
+        if 'nw_dst' in attrs and not 'arp_tpa' in attrs:
+            attrs['arp_tpa'] = attrs['nw_dst']
+            del attrs['nw_dst']
+
     for key, value in attrs.items():
         if key in convert:
             value = convert[key](value)
         if key in match_append:
             if key == 'nw_src' or key == 'nw_dst' or \
                     key == 'ipv4_src' or key == 'ipv4_dst' or \
+                    key == 'arp_spa' or key == 'arp_tpa' or \
                     key == 'ipv6_src' or key == 'ipv6_dst':
                 # IP address
                 ip = value[0]
@@ -347,8 +393,13 @@ def match_to_str(ofmatch):
             ofproto_v1_3.OXM_OF_TCP_DST: 'tp_dst',
             ofproto_v1_3.OXM_OF_UDP_SRC: 'tp_src',
             ofproto_v1_3.OXM_OF_UDP_DST: 'tp_dst',
+            ofproto_v1_3.OXM_OF_MPLS_LABEL: 'mpls_label',
             ofproto_v1_3.OXM_OF_METADATA: 'metadata',
             ofproto_v1_3.OXM_OF_METADATA_W: 'metadata',
+            ofproto_v1_3.OXM_OF_ARP_SPA: 'arp_spa',
+            ofproto_v1_3.OXM_OF_ARP_TPA: 'arp_tpa',
+            ofproto_v1_3.OXM_OF_ARP_SPA_W: 'arp_spa',
+            ofproto_v1_3.OXM_OF_ARP_TPA_W: 'arp_tpa',
             ofproto_v1_3.OXM_OF_IPV6_SRC: 'ipv6_src',
             ofproto_v1_3.OXM_OF_IPV6_DST: 'ipv6_dst',
             ofproto_v1_3.OXM_OF_IPV6_SRC_W: 'ipv6_src',
@@ -359,7 +410,8 @@ def match_to_str(ofmatch):
         key = keys[match_field.header]
         if key == 'dl_src' or key == 'dl_dst':
             value = mac.haddr_to_str(match_field.value)
-        elif key == 'nw_src' or key == 'nw_dst':
+        elif key == 'nw_src' or key == 'nw_dst' or \
+                key == 'arp_spa' or key == 'arp_tpa':
             value = match_ip_to_str(match_field.value, match_field.mask)
         elif key == 'ipv6_src' or key == 'ipv6_dst':
             value = match_ipv6_to_str(match_field.value, match_field.mask)
@@ -529,6 +581,11 @@ def get_meter_stats(dp, waiters):
 
 
 def get_meter_features(dp, waiters):
+
+    ofp = dp.ofproto
+    type_convert = {ofp.OFPMBT_DROP: 'DROP',
+                    ofp.OFPMBT_DSCP_REMARK: 'DSCP_REMARK'}
+
     stats = dp.ofproto_parser.OFPMeterFeaturesStatsRequest(dp, 0)
     msgs = []
     send_stats_request(dp, stats, waiters, msgs)
@@ -536,8 +593,12 @@ def get_meter_features(dp, waiters):
     features = []
     for msg in msgs:
         for feature in msg.body:
+            band_types = []
+            for k, v in type_convert.items():
+                if (1 << k) & feature.band_types:
+                    band_types.append(v)
             f = {'max_meter': feature.max_meter,
-                 'band_types': feature.band_types,
+                 'band_types': band_types,
                  'max_band': feature.max_band,
                  'max_color': feature.max_color}
             features.append(f)
@@ -581,6 +642,126 @@ def get_meter_config(dp, waiters):
     return configs
 
 
+def get_group_stats(dp, waiters):
+    stats = dp.ofproto_parser.OFPGroupStatsRequest(
+        dp, 0, dp.ofproto.OFPG_ALL)
+    msgs = []
+    send_stats_request(dp, stats, waiters, msgs)
+
+    groups = []
+    for msg in msgs:
+        for stats in msg.body:
+            bucket_stats = []
+            for bucket_stat in stats.bucket_stats:
+                c = {'packet_count': bucket_stat.packet_count,
+                     'byte_count': bucket_stat.byte_count}
+                bucket_stats.append(c)
+            g = {'group_id': stats.group_id,
+                 'ref_count': stats.ref_count,
+                 'packet_count': stats.packet_count,
+                 'byte_count': stats.byte_count,
+                 'duration_sec': stats.duration_sec,
+                 'duration_nsec': stats.duration_nsec,
+                 'bucket_stats': bucket_stats}
+            groups.append(g)
+    groups = {str(dp.id): groups}
+    return groups
+
+
+def get_group_features(dp, waiters):
+
+    ofp = dp.ofproto
+    type_convert = {ofp.OFPGT_ALL: 'ALL',
+                    ofp.OFPGT_SELECT: 'SELECT',
+                    ofp.OFPGT_INDIRECT: 'INDIRECT',
+                    ofp.OFPGT_FF: 'FF'}
+    cap_convert = {ofp.OFPGFC_SELECT_WEIGHT: 'SELECT_WEIGHT',
+                   ofp.OFPGFC_SELECT_LIVENESS: 'SELECT_LIVENESS',
+                   ofp.OFPGFC_CHAINING: 'CHAINING',
+                   ofp.OFPGFC_CHAINING_CHECKS: 'CHAINING_CHCEKS'}
+    act_convert = {ofp.OFPAT_OUTPUT: 'OUTPUT',
+                   ofp.OFPAT_COPY_TTL_OUT: 'COPY_TTL_OUT',
+                   ofp.OFPAT_COPY_TTL_IN: 'COPY_TTL_IN',
+                   ofp.OFPAT_SET_MPLS_TTL: 'SET_MPLS_TTL',
+                   ofp.OFPAT_DEC_MPLS_TTL: 'DEC_MPLS_TTL',
+                   ofp.OFPAT_PUSH_VLAN: 'PUSH_VLAN',
+                   ofp.OFPAT_POP_VLAN: 'POP_VLAN',
+                   ofp.OFPAT_PUSH_MPLS: 'PUSH_MPLS',
+                   ofp.OFPAT_POP_MPLS: 'POP_MPLS',
+                   ofp.OFPAT_SET_QUEUE: 'SET_QUEUE',
+                   ofp.OFPAT_GROUP: 'GROUP',
+                   ofp.OFPAT_SET_NW_TTL: 'SET_NW_TTL',
+                   ofp.OFPAT_DEC_NW_TTL: 'DEC_NW_TTL',
+                   ofp.OFPAT_SET_FIELD: 'SET_FIELD',
+                   ofp.OFPAT_PUSH_PBB: 'PUSH_PBB',
+                   ofp.OFPAT_POP_PBB: 'POP_PBB'}
+
+    stats = dp.ofproto_parser.OFPGroupFeaturesStatsRequest(dp, 0)
+    msgs = []
+    send_stats_request(dp, stats, waiters, msgs)
+
+    features = []
+    for msg in msgs:
+        feature = msg.body
+        types = []
+        for k, v in type_convert.items():
+            if (1 << k) & feature.types:
+                types.append(v)
+        capabilities = []
+        for k, v in cap_convert.items():
+            if k & feature.capabilities:
+                capabilities.append(v)
+        max_groups = []
+        for k, v in type_convert.items():
+            max_groups.append({v: feature.max_groups[k]})
+        actions = []
+        for k1, v1 in type_convert.items():
+            acts = []
+            for k2, v2 in act_convert.items():
+                if (1 << k2) & feature.actions[k1]:
+                    acts.append(v2)
+            actions.append({v1: acts})
+        f = {'types': types,
+             'capabilities': capabilities,
+             'max_groups': max_groups,
+             'actions': actions}
+        features.append(f)
+    features = {str(dp.id): features}
+    return features
+
+
+def get_group_desc(dp, waiters):
+
+    type_convert = {dp.ofproto.OFPGT_ALL: 'ALL',
+                    dp.ofproto.OFPGT_SELECT: 'SELECT',
+                    dp.ofproto.OFPGT_INDIRECT: 'INDIRECT',
+                    dp.ofproto.OFPGT_FF: 'FF'}
+
+    stats = dp.ofproto_parser.OFPGroupDescStatsRequest(dp, 0)
+    msgs = []
+    send_stats_request(dp, stats, waiters, msgs)
+
+    descs = []
+    for msg in msgs:
+        for stats in msg.body:
+            buckets = []
+            for bucket in stats.buckets:
+                actions = []
+                for action in bucket.actions:
+                    actions.append(action_to_str(action))
+                b = {'weight': bucket.weight,
+                     'watch_port': bucket.watch_port,
+                     'watch_group': bucket.watch_group,
+                     'actions': actions}
+                buckets.append(b)
+            d = {'type': type_convert.get(stats.type),
+                 'group_id': stats.group_id,
+                 'buckets': buckets}
+            descs.append(d)
+    descs = {str(dp.id): descs}
+    return descs
+
+
 def mod_flow_entry(dp, flow, cmd):
     cookie = int(flow.get('cookie', 0))
     cookie_mask = int(flow.get('cookie_mask', 0))
@@ -643,6 +824,38 @@ def mod_meter_entry(dp, flow, cmd):
     dp.send_msg(meter_mod)
 
 
+def mod_group_entry(dp, group, cmd):
+
+    type_convert = {'ALL': dp.ofproto.OFPGT_ALL,
+                    'SELECT': dp.ofproto.OFPGT_SELECT,
+                    'INDIRECT': dp.ofproto.OFPGT_INDIRECT,
+                    'FF': dp.ofproto.OFPGT_FF}
+
+    type_ = type_convert.get(group.get('type'))
+    if not type_:
+        LOG.debug('Unknown type: %s', group.get('type'))
+
+    group_id = int(group.get('group_id', 0))
+
+    buckets = []
+    for bucket in group.get('buckets', []):
+        weight = int(bucket.get('weight', 0))
+        watch_port = int(bucket.get('watch_port', dp.ofproto.OFPP_ANY))
+        watch_group = int(bucket.get('watch_group', dp.ofproto.OFPG_ANY))
+        actions = []
+        for dic in bucket.get('actions', []):
+            action = to_action(dp, dic)
+            if action is not None:
+                actions.append(action)
+        buckets.append(dp.ofproto_parser.OFPBucket(
+            weight, watch_port, watch_group, actions))
+
+    group_mod = dp.ofproto_parser.OFPGroupMod(
+        dp, cmd, type_, group_id, buckets)
+
+    dp.send_msg(group_mod)
+
+
 def send_experimenter(dp, exp):
     experimenter = exp.get('experimenter', 0)
     exp_type = exp.get('exp_type', 0)
