@@ -1,5 +1,5 @@
-# Copyright (C) 2013 Nippon Telegraph and Telephone Corporation.
-# Copyright (C) 2013 YAMAMOTO Takashi <yamamoto at valinux co jp>
+# Copyright (C) 2013,2014 Nippon Telegraph and Telephone Corporation.
+# Copyright (C) 2013,2014 YAMAMOTO Takashi <yamamoto at valinux co jp>
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -20,10 +20,7 @@ RFC 4271 BGP-4
 
 # todo
 # - notify data
-# - notify subcode constants
-# - RFC 3107 Carrying Label Information in BGP-4
 # - RFC 4364 BGP/MPLS IP Virtual Private Networks (VPNs)
-# - RFC 4486 Subcodes for BGP Cease Notification Message
 
 import abc
 import struct
@@ -34,6 +31,9 @@ from ryu.lib.packet import packet_base
 from ryu.lib.packet import stream_parser
 from ryu.lib import addrconv
 
+import safi
+import afi
+
 
 BGP_MSG_OPEN = 1
 BGP_MSG_UPDATE = 2
@@ -41,7 +41,10 @@ BGP_MSG_NOTIFICATION = 3
 BGP_MSG_KEEPALIVE = 4
 BGP_MSG_ROUTE_REFRESH = 5  # RFC 2918
 
-# RFC 4271 4.5.
+# NOTIFICATION Error Code and SubCode
+# Note: 0 is a valid SubCode.  (Unspecific)
+
+# NOTIFICATION Error Code  RFC 4271 4.5.
 BGP_ERROR_MESSAGE_HEADER_ERROR = 1
 BGP_ERROR_OPEN_MESSAGE_ERROR = 2
 BGP_ERROR_UPDATE_MESSAGE_ERROR = 3
@@ -49,6 +52,42 @@ BGP_ERROR_HOLD_TIMER_EXPIRED = 4
 BGP_ERROR_FSM_ERROR = 5
 BGP_ERROR_CEASE = 6
 
+# NOTIFICATION Error Subcode for BGP_ERROR_MESSAGE_HEADER_ERROR
+BGP_ERROR_SUB_CONNECTION_NOT_SYNCHRONIZED = 1
+BGP_ERROR_SUB_BAD_MESSAGE_LENGTH = 2  # Data: the erroneous Length field
+BGP_ERROR_SUB_BAD_MESSAGE_TYPE = 3  # Data: the erroneous Type field
+
+# NOTIFICATION Error Subcode for BGP_ERROR_OPEN_MESSAGE_ERROR
+BGP_ERROR_SUB_UNSUPPORTED_VERSION_NUMBER = 1  # Data: 2 octet version number
+BGP_ERROR_SUB_BAD_PEER_AS = 2
+BGP_ERROR_SUB_BAD_BGP_IDENTIFIER = 3
+BGP_ERROR_SUB_UNSUPPORTED_OPTIONAL_PARAMETER = 4
+# 5 is deprecated RFC 1771 Authentication Failure
+BGP_ERROR_SUB_UNACCEPTABLE_HOLD_TIME = 6
+
+# NOTIFICATION Error Subcode for BGP_ERROR_UPDATE_MESSAGE_ERROR
+BGP_ERROR_SUB_MALFORMED_ATTRIBUTE_LIST = 1
+BGP_ERROR_SUB_UNRECOGNIZED_WELL_KNOWN_ATTRIBUTE = 2  # Data: type of the attr
+BGP_ERROR_SUB_MISSING_WELL_KNOWN_ATTRIBUTE = 3  # Data: ditto
+BGP_ERROR_SUB_ATTRIBUTE_FLAGS_ERROR = 4  # Data: the attr (type, len, value)
+BGP_ERROR_SUB_ATTRIBUTE_LENGTH_ERROR = 5  # Data: ditto
+BGP_ERROR_SUB_INVALID_ORIGIN_ATTRIBUTE = 6  # Data: ditto
+# 7 is deprecated RFC 1771 AS Routing Loop
+BGP_ERROR_SUB_INVALID_NEXT_HOP_ATTRIBUTE = 8  # Data: ditto
+BGP_ERROR_SUB_OPTIONAL_ATTRIBUTE_ERROR = 9  # Data: ditto
+BGP_ERROR_SUB_INVALID_NETWORK_FIELD = 10
+BGP_ERROR_SUB_MALFORMED_AS_PATH = 11
+
+# NOTIFICATION Error Subcode for BGP_ERROR_CEASE  (RFC 4486)
+BGP_ERROR_SUB_MAXIMUM_NUMBER_OF_PREFIXES_REACHED = 1  # Data: optional
+BGP_ERROR_SUB_ADMINISTRATIVE_SHUTDOWN = 2
+BGP_ERROR_SUB_PEER_DECONFIGURED = 3
+BGP_ERROR_SUB_ADMINISTRATIVE_RESET = 4
+BGP_ERROR_SUB_CONNECTION_RESET = 5
+BGP_ERROR_SUB_OTHER_CONFIGURATION_CHANGE = 6
+BGP_ERROR_SUB_CONNECTION_COLLISION_RESOLUTION = 7
+BGP_ERROR_SUB_OUT_OF_RESOURCES = 8
+
 _VERSION = 4
 _MARKER = 16 * '\xff'
 
@@ -56,7 +95,10 @@ BGP_OPT_CAPABILITY = 2  # RFC 5492
 
 BGP_CAP_MULTIPROTOCOL = 1  # RFC 4760
 BGP_CAP_ROUTE_REFRESH = 2  # RFC 2918
+BGP_CAP_CARRYING_LABEL_INFO = 4  # RFC 3107
 BGP_CAP_FOUR_OCTET_AS_NUMBER = 65  # RFC 4893
+BGP_CAP_ENHANCED_ROUTE_REFRESH = 70  # https://tools.ietf.org/html/\
+                               # draft-ietf-idr-bgp-enhanced-route-refresh-05
 
 BGP_ATTR_FLAG_OPTIONAL = 1 << 7
 BGP_ATTR_FLAG_TRANSITIVE = 1 << 6
@@ -85,6 +127,7 @@ BGP_COMMUNITY_NO_ADVERTISE = 0xffffff02
 BGP_COMMUNITY_NO_EXPORT_SUBCONFED = 0xffffff03
 
 # RFC 4360
+# The low-order octet of Type field (subtype)
 BGP_EXTENDED_COMMUNITY_ROUTE_TARGET = 0x02
 BGP_EXTENDED_COMMUNITY_ROUTE_ORIGIN = 0x03
 
@@ -98,26 +141,34 @@ class _AddrPrefix(StringifyMixin):
     __metaclass__ = abc.ABCMeta
     _PACK_STR = '!B'  # length
 
-    def __init__(self, length, addr):
+    def __init__(self, length, addr, prefixes=None):
+        # length is on-wire bit length of prefixes+addr.
+        assert prefixes != ()
+        if isinstance(addr, tuple):
+            # for _AddrPrefix.parser
+            # also for _VPNAddrPrefix.__init__ etc
+            (addr,) = addr
         self.length = length
+        if prefixes:
+            addr = prefixes + (addr,)
         self.addr = addr
 
     @staticmethod
     @abc.abstractmethod
     def _to_bin(addr):
-        return addr
+        pass
 
     @staticmethod
     @abc.abstractmethod
-    def _to_text(addr):
-        return addr
+    def _from_bin(addr):
+        pass
 
     @classmethod
     def parser(cls, buf):
         (length, ) = struct.unpack_from(cls._PACK_STR, buffer(buf))
         rest = buf[struct.calcsize(cls._PACK_STR):]
         byte_length = (length + 7) / 8
-        addr = cls._to_text(rest[:byte_length])
+        addr = cls._from_bin(rest[:byte_length])
         rest = rest[byte_length:]
         return cls(length=length, addr=addr), rest
 
@@ -133,7 +184,7 @@ class _AddrPrefix(StringifyMixin):
             mask = 0xff00 >> (self.length % 8)
             last_byte = chr(ord(bin_addr[byte_length - 1]) & mask)
             bin_addr = bin_addr[:byte_length - 1] + last_byte
-        self.addr = self._to_text(bin_addr)
+        self.addr = self._from_bin(bin_addr)
 
         buf = bytearray()
         msg_pack_into(self._PACK_STR, buf, 0, self.length)
@@ -146,7 +197,76 @@ class _BinAddrPrefix(_AddrPrefix):
         return addr
 
     @staticmethod
-    def _to_text(addr):
+    def _from_bin(addr):
+        return addr
+
+
+class _LabelledAddrPrefix(_AddrPrefix):
+    _LABEL_PACK_STR = '!3B'
+
+    def __init__(self, length, addr, labels=[], **kwargs):
+        assert isinstance(labels, list)
+        is_tuple = isinstance(addr, tuple)
+        if is_tuple:
+            # for _AddrPrefix.parser
+            assert not labels
+            labels = addr[0]
+            addr = addr[1:]
+        else:
+            length += struct.calcsize(self._LABEL_PACK_STR) * 8 * len(labels)
+        assert length > struct.calcsize(self._LABEL_PACK_STR) * 8 * len(labels)
+        prefixes = (labels,)
+        super(_LabelledAddrPrefix, self).__init__(prefixes=prefixes,
+                                                  length=length,
+                                                  addr=addr,
+                                                  **kwargs)
+
+    @classmethod
+    def _label_to_bin(cls, label):
+        buf = bytearray()
+        msg_pack_into(cls._LABEL_PACK_STR, buf, 0,
+                      (label & 0xff0000) >> 16,
+                      (label & 0x00ff00) >> 8,
+                      (label & 0x0000ff) >> 0)
+        return buf
+
+    @classmethod
+    def _label_from_bin(cls, bin):
+        (b1, b2, b3) = struct.unpack_from(cls._LABEL_PACK_STR, buffer(bin))
+        rest = bin[struct.calcsize(cls._LABEL_PACK_STR):]
+        return (b1 << 16) | (b2 << 8) | b3, rest
+
+    @classmethod
+    def _to_bin(cls, addr):
+        labels = addr[0]
+        rest = addr[1:]
+        labels = map(lambda x: x << 4, labels)
+        if labels:
+            labels[-1] |= 1  # bottom of stack
+        bin_labels = map(cls._label_to_bin, labels)
+        return bytes(reduce(lambda x, y: x + y, bin_labels,
+                            bytearray()) + cls._prefix_to_bin(rest))
+
+    @classmethod
+    def _from_bin(cls, addr):
+        rest = addr
+        labels = []
+        while True:
+            (label, rest) = cls._label_from_bin(rest)
+            labels.append(label >> 4)
+            if label & 1:  # bottom of stack
+                break
+        return (labels,) + cls._prefix_from_bin(rest)
+
+
+class _UnlabelledAddrPrefix(_AddrPrefix):
+    @classmethod
+    def _to_bin(cls, addr):
+        return cls._prefix_to_bin((addr,))
+
+    @classmethod
+    def _from_bin(cls, binaddr):
+        (addr,) = cls._prefix_from_bin(binaddr)
         return addr
 
 
@@ -158,12 +278,68 @@ class _IPAddrPrefix(_AddrPrefix):
     }
 
     @staticmethod
-    def _to_bin(addr):
+    def _prefix_to_bin(addr):
+        (addr,) = addr
         return addrconv.ipv4.text_to_bin(addr)
 
     @staticmethod
-    def _to_text(addr):
-        return addrconv.ipv4.bin_to_text(pad(addr, 4))
+    def _prefix_from_bin(addr):
+        return (addrconv.ipv4.bin_to_text(pad(addr, 4)),)
+
+
+class _VPNAddrPrefix(_AddrPrefix):
+    _RD_PACK_STR = '!Q'
+
+    def __init__(self, length, addr, prefixes=(), route_dist=0):
+        if isinstance(addr, tuple):
+            # for _AddrPrefix.parser
+            assert not route_dist
+            assert length > struct.calcsize(self._RD_PACK_STR) * 8
+            route_dist = addr[0]
+            addr = addr[1:]
+        else:
+            length += struct.calcsize(self._RD_PACK_STR) * 8
+        prefixes = prefixes + (route_dist,)
+        super(_VPNAddrPrefix, self).__init__(prefixes=prefixes,
+                                             length=length,
+                                             addr=addr)
+
+    @classmethod
+    def _prefix_to_bin(cls, addr):
+        rd = addr[0]
+        rest = addr[1:]
+        binrd = bytearray()
+        msg_pack_into(cls._RD_PACK_STR, binrd, 0, rd)
+        return binrd + super(_VPNAddrPrefix, cls)._prefix_to_bin(rest)
+
+    @classmethod
+    def _prefix_from_bin(cls, binaddr):
+        binrd = binaddr[:8]
+        binrest = binaddr[8:]
+        (rd,) = struct.unpack_from(cls._RD_PACK_STR, buffer(binrd))
+        return (rd,) + super(_VPNAddrPrefix, cls)._prefix_from_bin(binrest)
+
+
+class LabelledVPNIPAddrPrefix(_LabelledAddrPrefix, _VPNAddrPrefix,
+                              _IPAddrPrefix):
+    pass
+
+
+class IPAddrPrefix(_UnlabelledAddrPrefix, _IPAddrPrefix):
+    pass
+
+
+_ADDR_CLASSES = {
+    (afi.IP, safi.UNICAST): IPAddrPrefix,
+    (afi.IP, safi.MPLS_VPN): LabelledVPNIPAddrPrefix,
+}
+
+
+def _get_addr_class(afi, safi):
+    try:
+        return _ADDR_CLASSES[(afi, safi)]
+    except KeyError:
+        return _BinAddrPrefix
 
 
 class _Value(object):
@@ -315,6 +491,15 @@ class _OptParamCapability(_OptParam, _TypeDisp):
         return buf + cap_value
 
 
+class _OptParamEmptyCapability(_OptParamCapability):
+    @classmethod
+    def parse_cap_value(cls, buf):
+        return {}
+
+    def serialize_cap_value(self):
+        return bytearray()
+
+
 @_OptParamCapability.register_unknown_type()
 class BGPOptParamCapabilityUnknown(_OptParamCapability):
     @classmethod
@@ -326,13 +511,8 @@ class BGPOptParamCapabilityUnknown(_OptParamCapability):
 
 
 @_OptParamCapability.register_type(BGP_CAP_ROUTE_REFRESH)
-class BGPOptParamCapabilityRouteRefresh(_OptParamCapability):
-    @classmethod
-    def parse_cap_value(cls, buf):
-        return {}
-
-    def serialize_cap_value(self):
-        return bytearray()
+class BGPOptParamCapabilityRouteRefresh(_OptParamEmptyCapability):
+    pass
 
 
 @_OptParamCapability.register_type(BGP_CAP_FOUR_OCTET_AS_NUMBER)
@@ -384,7 +564,12 @@ class BGPOptParamCapabilityMultiprotocol(_OptParamCapability):
         return buf
 
 
-class BGPWithdrawnRoute(_IPAddrPrefix):
+@_OptParamCapability.register_type(BGP_CAP_CARRYING_LABEL_INFO)
+class BGPOptParamCapabilityCarryingLabelInfo(_OptParamEmptyCapability):
+    pass
+
+
+class BGPWithdrawnRoute(IPAddrPrefix):
     pass
 
 
@@ -835,6 +1020,9 @@ class BGPPathAttributeMpReachNLRI(_PathAttribute):
         self.next_hop = next_hop
         self.reserved = reserved
         self.nlri = nlri
+        addr_cls = _get_addr_class(afi, safi)
+        for i in nlri:
+            assert isinstance(i, addr_cls)
 
     @classmethod
     def parse_value(cls, buf):
@@ -845,9 +1033,10 @@ class BGPPathAttributeMpReachNLRI(_PathAttribute):
         rest = rest[next_hop_len:]
         reserved = rest[:1]
         binnlri = rest[1:]
+        addr_cls = _get_addr_class(afi, safi)
         nlri = []
         while binnlri:
-            n, binnlri = _BinAddrPrefix.parser(binnlri)
+            n, binnlri = addr_cls.parser(binnlri)
             nlri.append(n)
         return {
             'afi': afi,
@@ -889,14 +1078,18 @@ class BGPPathAttributeMpUnreachNLRI(_PathAttribute):
         self.afi = afi
         self.safi = safi
         self.withdrawn_routes = withdrawn_routes
+        addr_cls = _get_addr_class(afi, safi)
+        for i in withdrawn_routes:
+            assert isinstance(i, addr_cls)
 
     @classmethod
     def parse_value(cls, buf):
         (afi, safi,) = struct.unpack_from(cls._VALUE_PACK_STR, buffer(buf))
         binnlri = buf[struct.calcsize(cls._VALUE_PACK_STR):]
+        addr_cls = _get_addr_class(afi, safi)
         nlri = []
         while binnlri:
-            n, binnlri = _BinAddrPrefix.parser(binnlri)
+            n, binnlri = addr_cls.parser(binnlri)
             nlri.append(n)
         return {
             'afi': afi,
@@ -914,7 +1107,7 @@ class BGPPathAttributeMpUnreachNLRI(_PathAttribute):
         return buf
 
 
-class BGPNLRI(_IPAddrPrefix):
+class BGPNLRI(IPAddrPrefix):
     pass
 
 
