@@ -21,6 +21,7 @@ import sys
 import functools
 
 from ryu import exception
+from ryu import utils
 from ryu.lib import stringify
 
 from . import ofproto_common
@@ -51,19 +52,55 @@ def msg(datapath, version, msg_type, msg_len, xid, buf):
     if msg_parser is None:
         raise exception.OFPUnknownVersion(version=version)
 
-    return msg_parser(datapath, version, msg_type, msg_len, xid, buf)
+    try:
+        return msg_parser(datapath, version, msg_type, msg_len, xid, buf)
+    except:
+        LOG.exception(
+            'Encounter an error during parsing OpenFlow packet from switch.'
+            'This implies switch sending a malfold OpenFlow packet.'
+            'version 0x%02x msg_type %d msg_len %d xid %d buf %s',
+            version, msg_type, msg_len, xid, utils.bytearray_to_hex(buf))
+        return None
 
 
 def create_list_of_base_attributes(f):
     @functools.wraps(f)
     def wrapper(self, *args, **kwargs):
         ret = f(self, *args, **kwargs)
-        self._base_attributes = set(dir(self))
+        cls = self.__class__
+        # hasattr(cls, '_base_attributes') doesn't work because super class
+        # may already have the attribute.
+        if '_base_attributes' not in cls.__dict__:
+            cls._base_attributes = set(dir(self))
         return ret
     return wrapper
 
 
 def ofp_msg_from_jsondict(dp, jsondict):
+    """
+    This function instanticates an appropriate OpenFlow message class
+    from the given JSON style dictionary.
+    The objects created by following two code fragments are equivalent.
+
+    Code A::
+
+        jsonstr = '{ "OFPSetConfig": { "flags": 0, "miss_send_len": 128 } }'
+        jsondict = json.loads(jsonstr)
+        o = ofp_msg_from_jsondict(dp, jsondict)
+
+    Code B::
+
+        o = dp.ofproto_parser.OFPSetConfig(flags=0, miss_send_len=128)
+
+    This function takes the following arguments.
+
+    ======== =======================================
+    Argument Description
+    ======== =======================================
+    dp       An instance of ryu.controller.Datapath.
+    jsondict A JSON style dict.
+    ======== =======================================
+    """
     parser = dp.ofproto_parser
     assert len(jsondict) == 1
     for k, v in jsondict.iteritems():
@@ -73,7 +110,7 @@ def ofp_msg_from_jsondict(dp, jsondict):
 
 
 class StringifyMixin(stringify.StringifyMixin):
-    _class_prefixes = ["OFP", "MT"]
+    _class_prefixes = ["OFP", "ONF", "MT"]
 
     @classmethod
     def cls_from_jsondict_key(cls, k):
@@ -83,6 +120,23 @@ class StringifyMixin(stringify.StringifyMixin):
 
 
 class MsgBase(StringifyMixin):
+    """
+    This is a base class for OpenFlow message classes.
+
+    An instance of this class has at least the following attributes.
+
+    ========= ==============================
+    Attribute Description
+    ========= ==============================
+    datapath  A ryu.controller.controller.Datapath instance for this message
+    version   OpenFlow protocol version
+    msg_type  Type of OpenFlow message
+    msg_len   Length of the message
+    xid       Transaction id
+    buf       Raw data
+    ========= ==============================
+    """
+
     @create_list_of_base_attributes
     def __init__(self, datapath):
         super(MsgBase, self).__init__()
@@ -122,10 +176,6 @@ class MsgBase(StringifyMixin):
         return msg_
 
     def _serialize_pre(self):
-        assert self.version is None
-        assert self.msg_type is None
-        assert self.buf is None
-
         self.version = self.datapath.ofproto.OFP_VERSION
         self.msg_type = self.cls_msg_type
         self.buf = bytearray(self.datapath.ofproto.OFP_HEADER_SIZE)
@@ -134,7 +184,6 @@ class MsgBase(StringifyMixin):
         # buffer length is determined after trailing data is formated.
         assert self.version is not None
         assert self.msg_type is not None
-        assert self.msg_len is None
         assert self.buf is not None
         assert len(self.buf) >= self.datapath.ofproto.OFP_HEADER_SIZE
 
